"""EVTX parsing tests."""
import unittest
import os

from libchickadee.parsers.evtx import EVTXParser

__author__ = 'Chapin Bryce'
__date__ = 20200710
__license__ = 'MIT Copyright 2020 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''


class EVTXParserTestCase(unittest.TestCase):
    """EVTX parsing tests."""
    def setUp(self):
        """Test Config."""
        self.test_data_ips = {'127.0.0.1': 6}
        self.parser = EVTXParser(ignore_bogon=False)
        self.test_data_dir = os.path.join(
            os.path.dirname(__file__), 'test_data')

    def test_ip_extraction_evtx(self):
        """Extraction test."""
        self.parser.parse_file(os.path.join(self.test_data_dir, 'System2.evtx'))
        self.assertEqual(self.test_data_ips, self.parser.ips)


if __name__ == '__main__':
    unittest.main()



"""
Libchickadee Overview
=====================

Yet another GeoIP resolution tool.

Libchickadee is a library consisting of tools to extract IP addresses from
files, resolve the IP address information using third-party data sources, and
present the resolutions to the end-user in an easy to consume manner. While the
library consists all of these components, it is primarily leveraged by a single
command line utility, ``chickadee``. The library components can be called
outside of the ``chickadee`` utility and this documentation provides
information on using the library in addition to the command line tool.

.. _installation:

Installation
------------

You may install Chickadee on your platform using ``pip install chickadee`` (you
may need to use ``pip3`` depending on your system configuration).
**Please ensure you are using Python 3**

You may also install via the source code as detailed below.

macOS and Linux
^^^^^^^^^^^^^^^

Requirements:

* Python 3+, installed on your path
* Virtualenv (``pip3 install virtualenv``)

#. Clone the git repo: ``git clone https://github.com/chapinb/chickadee.git``
#. Create your virtual environment ``virtualenv -p python3 venv3`` and
   activate it (``source venv3/bin/activate``)
#. Install dependencies: ``pip install .``
#. Run ``chickadee --help`` to get started.

Windows
^^^^^^^

Requirements:

* Python 3+, installed on your path
* Virtualenv (``pip.exe install virtualenv``)

#. Clone the git repo: ``git clone https://github.com/chapinb/chickadee.git``
#. Create your virtual environment ``virtualenv -p python3 venv3`` and activate
   it (``.\\venv3\\Scripts\\activate``)
#. Install dependencies: ``pip install .``
#. Run ``chickadee --help`` to get started.

Install the latest features without Git
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

If you would like to install the latest pre-release version, feel free to use
the above steps to clone the repository and install. You can also use pip
to directly install the latest master branch from GitHub by running:

```
pip install git+https://github.com/chapinb/chickadee.git
```

Contribution
------------

We appreciate contributions to the project! If you have an idea,
big or small, and want to lend a hand - feel free to follow the
steps in this guide to get started.

Looking for some inspiration? Check out the project `issues page
<https://github.com/chapinb/chickadee/issues>`_ for inspiration.
You are also welcome to contribute new ideas, features, or improvements.

We are happy to assist with any questions you have as you are working
on the project. Feel free to throw it as a new issue in GitHub. When
in doubt, just ask :) happy to provide guidance to those new to
contributing to projects!

The below steps outline the technical process for contributing to chickadee:

#. A `GitHub account <https://github.com/join>`_
#. Create a fork of the chickadee repository
#. Install git tools on your system
#. Clone your fork to your local system
#. Check out a new branch with a descriptive name.
   ie ``git checkout -b zipfile-support``
#. Install development dependencies in dev-requirements.txt
#. Make your modifications and write unit tests for new functionality
#. Submit a pull request to the main chickadee repository

A few tips to help get your new feature integrated smoothly:

* Ensure the whole project passes with ``flake8``. Avoid excluding linting
  warnings whenever possible.
* Run ``coverage run -m unittest discover`` and ensure all tests pass
  and all code files have at least 80% coverage. Avoid pragma statements
  whenever possible.
* Add documentation to your new functions/scripts and integrate into
  the overall project documentation. Build the documentation before
  submitting the pull request.
* As needed, update the argument and config file handling with help
  information to ensure users understand the purpose and functionality
  of the new capabilities.

When in doubt, reach out - we are happy to help with any of the above.

Usage & Examples
----------------

See :ref:`chickadee-usage` and :ref:`chickadee-examples`
for examples on using the ``chickadee`` command line tool.

Components
----------

Parsers
^^^^^^^

Collection of utilities for extracting IP addresses from different file
formats. The ``chickadee`` utility handles IP addresses provided via STDIN
and as argument parameters, though leverages these parsers to extract from
file content. The ``chickadee`` utility does iterate over directories to
expose files within folders recursively. With this, each parser needs to
accept a path to a file to read.

Resolvers
^^^^^^^^^

This is where the IP address is resolved by a third-party data source. Each
component is responsible for accepting one or more IP address and returning
resolution data. The third-party sources will not be distributed with this
library, each new resolver will have documentation to highlight the steps
required to prepare the data source for use by the resolver module.

The writers for libchickadee are also stored in this module.

Utilities
^^^^^^^^^

This includes the ``chickadee`` command line tool, used to interact with the
library from the command line.

"""

__author__ = 'Chapin Bryce'
__date__ = 20210314
__version__ = 20210314.0
__license__ = 'MIT Copyright 2021 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''



"""
ResolverBase
============

A base class for handling resolution of IP addresses.

This class includes elements common across resolver resolution data sources.
This includes common parameters, such as field names, and functions to handle
querying the data sources. Additionally, this includes writers for CSV and JSON
formats.

Module Documentation
--------------------
"""
import json
import csv

__author__ = 'Chapin Bryce'
__date__ = 20200107
__license__ = 'MIT Copyright 2020 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''


class ResolverBase:
    """Generic base class for use by other resolvers.

    Args:
        fields (list): Collection of fields to include in query.
        lang (str): Language for results.

    Returns:
        (ResolverBase)
    """
    def __init__(self):
        """Initialize class object and set defaults."""
        self.uri = None
        self.lang = 'en'
        self.supported_langs = []
        self.fields = []
        self.pbar = False  # Enable progress bars
        self.data = None

    def single(self):
        """Base method to handle single queries.

        Not implemented in base class.
        """
        raise NotImplementedError()

    def batch(self):
        """Base method to handle batch queries.

        Not implemented in base class."""
        raise NotImplementedError()

    def query(self, data):
        """Generic query handler to decide to use single or batch
            method for querying.

        Args:
            data (list, tuple, set, str): One or more IPs to resolve

        Returns:
            (list) List of collected records.

        Example:
            >>> records = ['1.1.1.1', '2.2.2.2']
            >>> resolver = ResolverBase()
            >>> resolved_data = resolver.query(records)
            >>> print(resolved_data)
            [
             {"query": "1.1.1.1", "country": "Australia", ...},
             {"query": "2.2.2.2", "country": "France", ...}
            ]
        """

        self.data = data
        if isinstance(data, (list, tuple, set)):
            return self.batch()
        if isinstance(data, str):
            return self.single()

        raise NotImplementedError()

    @staticmethod
    def defang_ioc(ioc):
        """Modify the display of IOCs to limit automated hyperlinking or access of unsafe resources

        Args:
            ioc (str): Content to scan for values to defang

        Return:
            (str): Defanged value.

        """
        return ioc.replace(".", "[.]")

    @staticmethod
    def write_csv(outfile, data, headers=None):
        """Writes a list of dictionaries to a CSV file.

        Arguments:
            outfile (str or file_obj): Path to output file
            data (list): List of dictionaries to write to file
            headers (list): Header row to use. If empty, will use the
                first dictionary in the `data` list.

        Returns:
            None

        Example:
            >>> records = [{'query': '1.1.1.1', 'count': 2}]
            >>> resolver = ResolverBase()
            >>> resolver.write_csv('test.csv', records, ['query', 'count'])

        """

        if not headers:
            # Use the first line of data
            headers = [str(x) for x in data[0].keys()]

        # Write rows individually to handle flattening complex objects. Will update headers with new fields
        rows_to_write, headers = ResolverBase.flatten_objects(data, headers)

        was_opened = False
        if isinstance(outfile, str):
            open_file = open(outfile, 'w', newline="")
            was_opened = True
        else:
            open_file = outfile

        # Write only provided headers, ignore others
        csvfile = csv.DictWriter(open_file, headers,
                                 extrasaction='ignore')
        csvfile.writeheader()

        csvfile.writerows(rows_to_write)

        if was_opened:
            open_file.close()

    @staticmethod
    def write_json(outfile, data, headers=None, lines=False):
        """Writes output in JSON format

        Args:
            outfile (str or file_obj): Path to or already open file
            data (list): List of dictionaries containing resolved data
            headers (list): List of column headers. Will use the first element of data if not present.
            lines (bool): Whether to export 1 dictionary object per line or
                a whole json object.

        Returns:
            None

        Examples:
            Generation of a JSON report

            >>> records = [{'query': '1.1.1.1', 'count': 2}]
            >>> resolver = ResolverBase()
            >>> resolver.write_json('test.json', records, ['query', 'count'])

            Generation of a JSON Lines report

            >>> recs = [{'query': '1.1.1.1', 'count': 2}]
            >>> resolver = ResolverBase()
            >>> resolver.write_json('out.json', recs, ['query', 'count'], True)

        """
        was_opened = False
        open_file = outfile
        if isinstance(outfile, str):
            open_file = open(outfile, 'w', newline="")
            was_opened = True

        if headers:
            data = ResolverBase.normalize_data_headers(data, headers)

        if lines:
            for entry in data:
                open_file.write(json.dumps(entry)+"\n")
        else:
            open_file.write(json.dumps(data))

        if was_opened:
            open_file.close()

    @staticmethod
    def normalize_data_headers(data, headers):
        """Filter content from response that is not requested in output

        Args:
            data (list): Response from API
            headers (list): List of user-provided headers to return

        Returns:
            (dict): Updated API response with limited headers
        """
        # Only include fields in headers
        # Include headers with no value if not present in original
        selected_data = []
        for x in data:
            d = {}
            for k, v in x.items():
                if k in headers:
                    d[k] = v
            for h in headers:
                if h not in d:
                    d[h] = None
            selected_data.append(d)
        return selected_data

    @staticmethod
    def flatten_objects(data, headers):
        """Flatten complex fields in to simple columns for CSV usage

        Args:
            data (list): List of dictionaries to write to file
            headers (list): Header row to use.

        Returns:
            (tuple, list): Tuple containing a list of rows to write, followed by a list of headers to use.
        """
        rows_to_write = []
        for raw_row in data:
            row = raw_row.copy()
            # Convert lists in to CSV friendly format
            for header in headers:
                ResolverBase._process_header(header, headers, raw_row, row)
            rows_to_write.append(row)
        return rows_to_write, headers

    @staticmethod
    def _process_header(header, headers, raw_row, row):
        """Extract list of headers in to a CSV format, flatten any nested dictionaries with dot notation

        Args:
            header (str): Field name
            headers (list): List of all fields
            raw_row (dict): Original row
            row (dict): Updated row

        Returns:
            None. Updates the `row` dictionary provided
        """
        if isinstance(raw_row.get(header, None), list):
            # Converts list of simple values (str, int, float, bool) to pipe delimited string
            row[header] = " | ".join(raw_row[header])
        elif isinstance(raw_row.get(header, None), dict):
            # For each object in a dictionary, add a new header and append to
            for key, value in raw_row[header].items():
                new_header = '{}.{}'.format(header, key)
                if new_header not in headers:
                    headers.append(new_header)
                row[new_header] = value



"""IP-API Resolver Tests."""
import unittest
from datetime import datetime
from unittest.mock import patch
import csv
import json
import os
from collections import OrderedDict

from libchickadee.resolvers.ipapi import Resolver

__author__ = 'Chapin Bryce'
__date__ = 20200114
__license__ = 'MIT Copyright 2020 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''


class MockResponse:
    """Generate mocked responses from the ip-api.com source."""
    def __init__(self, json_data, status_code, rl='100', ttl='1'):
        """Setup accessed objects."""
        self.run_count = 0
        self.json_data = json_data
        self.status_code = status_code
        self.headers = {'X-Rl': rl, 'X-Ttl': ttl}

    def json(self):
        """Return provided sample data."""
        return self.json_data


class IPAPITestCase(unittest.TestCase):
    """IP-API Resolver Tests."""
    def setUp(self):
        """Test config"""
        self.test_data_ips = [
            '10.0.1.2', '8.8.8.8', '2001:4860:4860::8888'
        ]
        self.expected_result = [
            {'query': '10.0.1.2'},

            {'as': 'AS15169 Google LLC', 'country': 'United States',
             'org': 'Level 3', 'proxy': False, 'query': '8.8.8.8'},

            {'as': 'AS15169 Google LLC', 'country': 'United States',
             'org': 'Google LLC', 'proxy': False,
             'query': '2001:4860:4860::8888'}
        ]
        self.resolver = Resolver(fields=['query', 'count', 'as',
                                         'country', 'org', 'proxy'])

    @patch("libchickadee.resolvers.ipapi.Resolver.single")
    def test_ipapi_resolve_query_single(self, mock_query):
        """Query Method Test"""
        for count, ip in enumerate(self.test_data_ips):
            mock_query.return_value = self.expected_result[count]
            data = self.resolver.query(ip)
            self.assertEqual(data, self.expected_result[count])

    @patch("libchickadee.resolvers.ipapi.Resolver.batch")
    def test_ipapi_resolve_query_batch(self, mock_query):
        """Batch Query Method Test"""
        mock_query.return_value = self.expected_result.copy()
        data = self.resolver.query(self.test_data_ips)
        res = [x for x in data]
        batch_result = []  # No reverse field
        for item in self.expected_result:
            batch_result.append(item)
        self.assertCountEqual(res, batch_result)

    @patch("libchickadee.resolvers.ipapi.requests.get")
    def test_ipapi_resolve_single(self, mock_query):
        """Single Query Method Test"""
        for count, ip in enumerate(self.test_data_ips):
            mock_query.return_value = MockResponse(json_data=self.expected_result[count], status_code=200)
            self.resolver.data = ip
            data = self.resolver.single()
            self.assertEqual(data, [self.expected_result[count]])

    @patch("libchickadee.resolvers.ipapi.requests.post")
    def test_ipapi_resolve_batch(self, mock_query):
        """Batch Query Method Test"""
        mock_query.return_value = MockResponse(json_data=self.expected_result, status_code=200)
        self.resolver.data = self.test_data_ips
        data = self.resolver.batch()
        res = [x for x in data]
        self.assertEqual(len(res), len(self.expected_result))

    @patch("libchickadee.resolvers.ipapi.Resolver.single")
    def test_ipapi_resolve_single_field(self, mock_query):
        """Single Query Method Test"""
        for count, ip in enumerate(self.test_data_ips):
            expected = {
                'query': self.expected_result[count].get('query', None),
                'country': self.expected_result[count].get('country', None),
                'as': self.expected_result[count].get('as', None),
            }

            mock_query.return_value = expected
            self.resolver.data = ip
            self.resolver.fields = ['query', 'country', 'as']
            data = self.resolver.single()

            self.assertEqual(data, expected)

    @patch("libchickadee.resolvers.ipapi.requests.get")
    @patch("libchickadee.resolvers.ipapi.requests.post")
    def test_ipapi_rate_limiting(self, mock_get, mock_post):
        """Validate the handling for rate limiting"""
        single = {
            "test_data": self.test_data_ips[1],
            "expected_data": [self.expected_result[1]],
            "mock_data": [
                MockResponse(json_data={}, status_code=429, rl='0', ttl='2'),
                MockResponse(json_data=self.expected_result[1], status_code=200, rl='0', ttl='0')
            ]
        }
        batch = {
            "test_data": self.test_data_ips,
            "expected_data": self.expected_result,
            "mock_data": [
                MockResponse(json_data={}, status_code=429, rl='0', ttl='2'),
                MockResponse(json_data=self.expected_result, status_code=200, rl='0', ttl='0')
            ]
        }
        for test in [single, batch]:
            test_ip = test['test_data']
            mock_get.side_effect = test['mock_data']
            mock_post.side_effect = test['mock_data']
            start_time = datetime.now()
            data = self.resolver.query(test_ip)
            delta = datetime.now() - start_time
            self.assertTrue(delta.total_seconds() > 1)
            self.assertEqual(data, test['expected_data'])


class WritersTestCase(unittest.TestCase):
    """Test writer functionality"""
    def setUp(self):
        """Configure test data"""
        self.data = [
            OrderedDict(**{"a": '1', "b": "2", "regionName": "Île-de-France"})
        ]
        self.testfile = "testfile"
        self.open_file = None

    def tearDown(self):
        """Clean up test data"""
        self.open_file.close()
        os.remove(self.testfile)

    def test_write_csv(self):
        """Test writing information to a CSV file."""
        Resolver.write_csv(self.testfile, self.data)
        self.open_file = open("testfile")
        read_data = next(csv.DictReader(self.open_file))
        self.assertDictEqual(
            self.data[0],
            read_data
        )

    def test_write_json(self):
        """Test writing information to a JSON file"""
        data = [
            {"a": '1', "b": "2", "regionName": "Île-de-France"}
        ]
        Resolver.write_json(self.testfile, self.data)
        self.open_file = open("testfile")
        read_data = json.load(self.open_file)
        self.assertEqual(
            data,
            read_data
        )

    def test_write_json_headers(self):
        """Test writing information to a JSON file with filtered headers"""
        Resolver.write_json(self.testfile, self.data, ['a', 'none'])
        self.open_file = open("testfile")
        read_data = json.load(self.open_file)
        rec = self.data[0]
        rec.pop("b")
        rec.pop("regionName")
        rec["none"] = None
        self.assertDictEqual(
            rec,
            read_data[0]
        )

    def test_write_json_lines(self):
        """Test writing information to a JSON file with one object per line"""
        Resolver.write_json(self.testfile, self.data, lines=True)
        self.open_file = open("testfile")
        read_data = json.load(self.open_file)
        self.assertDictEqual(
            self.data[0],
            read_data
        )


if __name__ == '__main__':
    unittest.main()



"""
Common Parser Utilities
=======================

Collection of input parser utilities to extract IP addresses.

This includes common regex patterns and utilities for extracting IP addresses
for resolution.

"""
import json
import os
import re
import sys

from netaddr import IPAddress

__author__ = 'Chapin Bryce'
__date__ = 20200107
__license__ = 'MIT Copyright 2020 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''

# FROM https://gist.github.com/dfee/6ed3a4b05cfe7a6faf40a2102408d5d8
IPV4SEG = r'(?:25[0-5]|(?:2[0-4]|1{0,1}[0-9]){0,1}[0-9])'
IPV4ADDR = r'(?:(?:' + IPV4SEG + r'\.){3,3}' + IPV4SEG + r')'

IPV6SEG = r'(?:(?:[0-9a-fA-F]){1,4})'
IPV6GROUPS = (
    # 1:2:3:4:5:6:7:8
    r'(?:' + IPV6SEG + r':){7,7}' + IPV6SEG,
    # 1:: | 1:2:3:4:5:6:7::
    r'(?:' + IPV6SEG + r':){1,7}:',
    # 1::8 | 1:2:3:4:5:6::8 | 1:2:3:4:5:6::8
    r'(?:' + IPV6SEG + r':){1,6}:' + IPV6SEG,
    # 1::7:8 | 1:2:3:4:5::7:8 | 1:2:3:4:5::8
    r'(?:' + IPV6SEG + r':){1,5}(?::' + IPV6SEG + r'){1,2}',
    # 1::6:7:8 | 1:2:3:4::6:7:8 | 1:2:3:4::8
    r'(?:' + IPV6SEG + r':){1,4}(?::' + IPV6SEG + r'){1,3}',
    # 1::5:6:7:8 | 1:2:3::5:6:7:8 | 1:2:3::8
    r'(?:' + IPV6SEG + r':){1,3}(?::' + IPV6SEG + r'){1,4}',
    # 1::4:5:6:7:8 | 1:2::4:5:6:7:8 | 1:2::8
    r'(?:' + IPV6SEG + r':){1,2}(?::' + IPV6SEG + r'){1,5}',
    # 1::3:4:5:6:7:8 | 1::3:4:5:6:7:8 | 1::8
    IPV6SEG + r':(?:(?::' + IPV6SEG + r'){1,6})',
    # ::2:3:4:5:6:7:8 | ::2:3:4:5:6:7:8 | ::8 | ::
    r':(?:(?::' + IPV6SEG + r'){1,7}|:)',
    # fe80::7:8%eth0 | fe80::7:8%1
    # (link-local IPv6 addresses with zone index)
    r'fe80:(?::' + IPV6SEG + r'){0,4}%[0-9a-zA-Z]{1,}',
    # ::255.255.255.255 | ::ffff:255.255.255.255 | ::ffff:0:255.255.255.255
    # (IPv4-mapped IPv6 addresses and IPv4-translated addresses)
    r'::(?:ffff(?::0{1,4}){0,1}:){0,1}[^\s:]' + IPV4ADDR,
    # 2001:db8:3:4::192.0.2.33  64:ff9b::192.0.2.33
    # (IPv4-Embedded IPv6 Address)
    r'(?:' + IPV6SEG + r':){1,4}:[^\s:]' + IPV4ADDR,
)

# Reverse rows for greedy match
IPV6ADDR = '|'.join('(?:{})'.format(g) for g in IPV6GROUPS[::-1])

IPv4Pattern = re.compile(IPV4ADDR)
IPv6Pattern = re.compile(IPV6ADDR)


def run_parser_from_cli(args, parser_obj):  # pragma: no cover
    """Allow a parser to run from the command line, both for testing and increased usability."""
    if os.path.isdir(args.path):
        for root, _, files in os.walk(args.path):
            for fentry in files:
                parser_obj.parse_file(os.path.join(root, fentry))
    else:
        parser_obj.parse_file(args.path)
    sys.stderr.write("{} unique IPs discovered, shown below with their frequency.\n".format(len(parser_obj.ips)))
    for ip, count in parser_obj.ips.items():
        print(json.dumps({"count": count, "ip": ip}))


class ParserBase(object):
    """Base class for parsers, containing common utilities."""
    def __init__(self, ignore_bogon=True):
        """Configure the parser and set default values."""
        self.ignore_bogon = ignore_bogon
        self.ips = {}

    def check_ips(self, data):
        """Check data for IP addresses. Results stored in ``self.ips``.

        Args:
            data (str): String to search for IP address content.

        Returns:
            None
        """
        for ipv4 in IPv4Pattern.findall(data):
            if self.ignore_bogon and self.is_bogon(ipv4):
                continue
            if ipv4 not in self.ips:
                self.ips[ipv4] = 0
            self.ips[ipv4] += 1
        for ipv6 in IPv6Pattern.findall(data):
            ipv6 = self.strip_ipv6(ipv6)
            if self.ignore_bogon and self.is_bogon(ipv6):
                continue
            if ipv6 not in self.ips:
                self.ips[ipv6] = 0
            self.ips[ipv6] += 1

    @staticmethod
    def strip_ipv6(ipv6_addr):
        """Isolate IPv6 Value containing a ``%`` symbol.

        Args:
            ipv6_addr (str): Raw IPv6 IP address to strip.

        Returns:
            (str): IP address base.
        """
        if '%' in ipv6_addr:
            ip, _ = ipv6_addr.split('%')
        else:
            ip = ipv6_addr
        return ip

    @staticmethod
    def is_bogon(ip_addr):
        """Identifies whether an IP address is a known BOGON.

        Args:
            ip_addr (str): Valid IP address to check.

        Returns:
            (bool): Whether or not the IP is a known BOGON address.
        """
        ip = IPAddress(ip_addr)
        return bool((ip.is_private() or ip.is_link_local() or
                     ip.is_reserved() or ip.is_multicast()))



"""
VirusTotal Resolver
===================

Resolver leveraging the VirusTotal JSON API.

This is a third-party data source that provides GeoIP, ASN, associated detected file samples,
detections on web hosted content, known resolutions, and other resolution information for IPv4 and IPv6 addresses.

Data Source Information
-----------------------

This data source is hosted at virustotal.com and requires an internet connection
to use. It offers a free API with registration and approval, with terms described
on their site. Please refer to the service's website for an authoritative source
on API specifications.

This documentation summarizes a point in time understanding of the data source
though since it is a third party service, it may change in a manner that breaks
this tool or causes this documentation to become inaccurate. In no way is
inclusion of a data source in libchickadee an endorsement of the data source.

**Data source documentation:** https://developers.virustotal.com/reference

Endpoints
^^^^^^^^^

The API supports a number of reports, though for IP addresses only allows
the submission of a single IP address per request. This means bulk lookups
will take at most 1 minute per 4 IP addresses, due to rate limiting.

* ``https://www.virustotal.com/vtapi/v2/ip-address/report?apikey=<apikey>&ip=<ip>``

Fields
^^^^^^

These fields are in no particular order.

* query
* count
* asn
* country
* subnet
* resolution_count
* detected_sample_count
* detected_samples
* undetected_sample_count
* undetected_samples
* detected_url_count
* detected_urls
* undetected_url_count
* undetected_urls
* status
* message

Limitations
^^^^^^^^^^^

This service has a free tier for non-commercial use, and is rate limited to:

* 4 requests per minute
* 5760 requests per day
* 172800 requests per month

A 204 response code indicates that you have exceeded your rate limit. The script will
sleep for 15 seconds if a 204 is returned.

Module Documentation
--------------------

"""
import collections
import functools
import logging
import operator
import time
from datetime import datetime

import requests
from tqdm import tqdm

from . import ResolverBase

logger = logging.getLogger(__name__)

__author__ = 'Chapin Bryce'
__date__ = 20200805
__license__ = 'MIT Copyright 2020 Chapin Bryce'
__desc__ = 'Resolver for VirusTotal'

FIELDS = [
    'query', 'count',
    'asn', 'country',
    'subnet',
    'resolution_count', 'detected_sample_count', 'undetected_sample_count',
    'detected_url_count', 'undetected_url_count',
    'status', 'message'
]

NON_DEFAULT_FIELDS = [
    "resolutions", "detected_samples", "undetected_samples", "detected_urls", "undetected_urls"
]


class ProResolver(ResolverBase):
    """VirusTotal authenticated API resolver. Requires an API key.
    Full API documentation here: https://developers.virustotal.com/reference"""
    def __init__(self, api_key, fields=None, lang="en"):
        """Initialize class object and configure default values.

        Args:
            api_key (str): API key to authenticate with
            fields (list): Fields to filter responses
            lang (str): Language to use in the API request
        """
        super().__init__()

        self.supported_langs = [
            "en"
        ]
        self.lang = 'en' if lang not in self.supported_langs else lang
        self.fields = FIELDS if not fields else fields
        self.uri = "https://www.virustotal.com/vtapi/v2/ip-address/report"
        self.api_key = api_key
        self.enable_sleep = True
        self.last_request = datetime.now()
        logger.debug("API key found")

    def sleeper(self):
        """Method to sleep operations for rate limiting.

        Will ensure that 4 requests per minute limit is not exceeded if a 204 encountered.

        Return:
            None
        """
        current_request = datetime.now()
        time_since_last_request = current_request - self.last_request
        if time_since_last_request.total_seconds() > 15:
            return
        time_to_sleep = (15-time_since_last_request.total_seconds()) + .25  # Add padding
        logger.info('Sleeping for {} seconds due to rate limiting.'.format(time_to_sleep))
        time.sleep(time_to_sleep)

    def batch(self):
        """Resolve multiple IP addresses.

        Due to API limitations, each IP address must be resolved individually.

        Returns:
            (list): List of resolved results
        """
        # Not supported
        all_ips = self.data
        records = []
        if self.pbar:
            all_ips = tqdm(self.data)
        for x in all_ips:
            self.data = x
            resp = self.single()
            if resp:
                records.append(resp[0])
        return records

    def single(self):
        """Gathers VirusTotal report data for a single IP address.

        Returns:
            (list): Report information
        """
        params = {
            'apikey': self.api_key,
            'ip': self.data
        }

        self.last_request = datetime.now()
        rdata = requests.get(
            self.uri, params=params
        )

        if rdata.status_code == 200:
            return [self.parse_vt_resp(self.data, rdata.json())]
        if rdata.status_code == 204:
            # Rate limit
            self.sleeper()
            # Try again
            return self.single()
        elif rdata.status_code == 400:
            logger.error("Incorrect request. Please check input data")
        elif rdata.status_code == 403:
            logger.error("Authorization error. Please check API key")
        else:
            logger.error("Unknown error occurred, status code {}, please report".format(rdata.status_code))

    def parse_vt_resp(self, query, vt_resp):
        """Transform the raw response from VirusTotal in to a dictionary easier for analysis

        Args:
            query (str): Queried value
            vt_resp (dict): Raw API response

        Returns:
            (dict): Formatted response for analysis
        """
        attributes = dict.fromkeys(self.fields.copy())
        attributes.update({
            "query": query,
            "asn": "",
            "continent": vt_resp.get("continent"),
            "country": vt_resp.get("country"),
            "subnet": vt_resp.get("network"),
            "status": vt_resp.get("response_code"),
            "message": vt_resp.get("IP address in dataset"),
            "whois": "",
        })

        # ASNs
        if vt_resp.get("asn"):
            attributes["asn"] = "AS{} {}".format(
                vt_resp.get("asn"), vt_resp.get("as_owner")
            )

        # Parse WhoIs
        self._extract_whois(attributes, vt_resp)

        # Parse Resolutions
        self._extract_resolutions(attributes, vt_resp)

        # Parse Detected Communicating, Download, and Referrer Samples
        self._extract_detected_samples(attributes, vt_resp)

        # Parse Undetected Communicating Samples
        self._extract_undetected_samples(attributes, vt_resp)

        # Parse Detected URLs
        self._extract_detected_urls(attributes, vt_resp)

        # Parse Undetected URLs
        self._extract_undetected_urls(attributes, vt_resp)

        return attributes

    @staticmethod
    def _extract_resolutions(attributes, vt_resp):
        """Extract information about associated domain resolutions

        Args:
            attributes (dict): Formatted data to present for analysis
            vt_resp (dict): Raw information from the VT API

        Returns:
            None. Updates attributes dictionary
        """
        attributes["resolution_count"] = len(vt_resp.get("resolutions", []))
        hostname_set = {x.get('hostname') for x in
                        vt_resp.get("resolutions", [])}
        attributes["resolutions"] = sorted(list(hostname_set))

    def _extract_undetected_urls(self, attributes, vt_resp):
        """Extract information about undetected URLs

        Args:
            attributes (dict): Formatted data to present for analysis
            vt_resp (dict): Raw information from the VT API

        Returns:
            None. Updates attributes dictionary
        """
        # * Get count
        # * Get all defanged URLs
        # The API does not expose this as a key/value object
        # but instead as a list where the elements are:
        # * 0: URL
        # * 1: SHA256 Hash value
        # * 2: Number of positive detections
        # * 3: Number of scanners
        # * 4: Time of last scan
        attributes["undetected_url_count"] = len(
            vt_resp.get("undetected_urls", []))
        detected_urls = {
            self.defang_ioc(x[0])
            for x in vt_resp.get('undetected_urls', [])}
        attributes["undetected_urls"] = sorted(list(detected_urls))

    def _extract_detected_urls(self, attributes, vt_resp):
        """Extract information about detected URLs

        Args:
            attributes (dict): Formatted data to present for analysis
            vt_resp (dict): Raw information from the VT API

        Returns:
            None. Updates attributes dictionary
        """
        # * Get count
        # * Get all defanged URLs
        attributes["detected_url_count"] = len(
            vt_resp.get("detected_urls", []))
        detected_urls = {
            self.defang_ioc(x.get('url')): x.get('positives')
            for x in vt_resp.get('detected_urls', [])}
        attributes["detected_urls"] = detected_urls

    @staticmethod
    def _extract_undetected_samples(attributes, vt_resp):
        """Extract information about undetected samples

        Args:
            attributes (dict): Formatted data to present for analysis
            vt_resp (dict): Raw information from the VT API

        Returns:
            None. Updates attributes dictionary
        """
        # * Get count
        # * Get all hashes
        attributes["undetected_sample_count"] = len(vt_resp.get("undetected_communicating_samples", [])) + \
                                                len(vt_resp.get("undetected_downloaded_samples", [])) + len(
            vt_resp.get("undetected_referrer_samples", []))
        undetected_samples = {
            x.get('sha256') for x in
            vt_resp.get('undetected_communicating_samples', [])}
        undetected_samples = undetected_samples.union({
            x.get('sha256') for x in
            vt_resp.get('undetected_downloaded_samples', [])})
        undetected_samples = undetected_samples.union({
            x.get('sha256') for x in
            vt_resp.get('undetected_referrer_samples', [])})
        attributes["undetected_samples"] = sorted(list(undetected_samples))

    @staticmethod
    def _extract_detected_samples(attributes, vt_resp):
        """Extract information about detected samples

        Args:
            attributes (dict): Formatted data to present for analysis
            vt_resp (dict): Raw information from the VT API

        Returns:
            None. Updates attributes dictionary
        """
        # * Get count
        # * Get all hashes
        attributes["detected_sample_count"] = len(vt_resp.get("detected_communicating_samples", [])) + \
                                              len(vt_resp.get("detected_downloaded_samples", [])) + len(
            vt_resp.get("detected_referrer_samples", []))
        detected_communicating_samples = {
            x.get('sha256'): x.get('positives')
            for x in vt_resp.get('detected_communicating_samples', [])}
        detected_downloaded_samples = {
            x.get('sha256'): x.get('positives')
            for x in vt_resp.get('detected_downloaded_samples', [])}
        detected_referrer_samples = {
            x.get('sha256'): x.get('positives')
            for x in vt_resp.get('detected_referrer_samples', [])}
        # Sum up the counts across categories for the same samples
        attributes["detected_samples"] = dict(
            functools.reduce(
                operator.add, map(
                    collections.Counter, [detected_communicating_samples,
                                          detected_downloaded_samples,
                                          detected_referrer_samples])))

    @staticmethod
    def _extract_whois(attributes, vt_resp):
        """Extract whois data from the VirusTotal API response

        Args:
            attributes (dict): Formatted data to present for analysis
            vt_resp (dict): Raw information from the VT API

        Returns:
            None. Updates attributes dictionary
        """
        # OtherRemarks holds info not stored in key/value config
        whois = {"OtherRemarks": ""}
        for line in vt_resp.get('whois', "").split("\n"):
            if len(line) == 0:
                continue
            if ':' not in line:
                whois["OtherRemarks"] += line + "\n"
                continue
            split_line = line.split(":", 1)
            whois[split_line[0].strip()] = split_line[1].strip()
        attributes["whois"] = whois



"""
IP-API Resolver
===============

Resolver leveraging the ip-api.com JSON API.

This is a third-party data source that provides GeoIP, ASN, Organization,
ISP, and other resolution information for IPv4 and IPv6 addresses.

Data Source Information
-----------------------

This data source is hosted at ip-api.com and requires an internet connection
to use. It offers a free API, with terms described on their site. Please refer
to the service's website for an authoritative source on API specifications.
This documentation summarizes a point in time understanding of the data source
though since it is a third party service, it may change in a manner that breaks
this tool or causes this documentation to become inaccurate. In no way is
inclusion of a data source in libchickadee an endorsement of the data source.

**Data source documentation:** https://ip-api.com/docs

Endpoints
^^^^^^^^^

The API supports a number of formats, though this resolver implementation
leverages the JSON APIs. This includes both the ``single`` and ``batch`` APIs
depending on the number of IP addresses requested for resolution.

* ``http://ip-api.com/json/{query}``
* ``http://ip-api.com/batch``

Fields
^^^^^^

These fields are in no particular order.

* status
* message
* continent
* continentCode
* country
* countryCode
* region
* regionName
* city
* district
* zip
* lat
* lon
* timezone
* currency
* isp
* org
* as
* asname
* mobile
* proxy
* hosting
* query

Limitations
^^^^^^^^^^^

This service has a free tier for non-commercial use, and is rate limited to
15 requests per minute. The returned HTTP header ``X-Rl`` contains the number
of requests remaining in the current rate limit window. ``X-Ttl`` contains the
seconds until the limit is reset.

The professional service is supported by chickadee and allows the execution
of unlimited requests, commercial use, and https endpoints. More details are
available on the data source website.

Module Documentation
--------------------

"""
import logging
import time
from datetime import datetime, timedelta

import requests
from tqdm import trange

from . import ResolverBase

logger = logging.getLogger(__name__)

__author__ = 'Chapin Bryce'
__date__ = 20200114
__license__ = 'MIT Copyright 2020 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''

FIELDS = [  # Ordered list of fields to gather
    'query', 'count',
    'as', 'org', 'isp',
    'continent', 'country', 'regionName', 'city',
    'district', 'zip',
    'mobile', 'proxy', 'reverse',
    'lat', 'lon', 'timezone',
    'status', 'message'
]


class Resolver(ResolverBase):
    """Class to handle ip-api.com API queries for IP addresses.

    Sets endpoint to free API, confirms support for requested language,
    configures rate limit sleep timers.

    Args:
        fields (list): Collection of fields to request in resolution.
        lang (str): Language for returned results.
    """
    def __init__(self, fields=None, lang='en'):
        """Initialize class object and configure default values."""
        self.supported_langs = [
            'en', 'de', 'es', 'pt-BR', 'fr', 'ja', 'zh-CN', 'ru'
        ]
        super().__init__()

        self.lang = 'en' if lang not in self.supported_langs else lang
        self.fields = [] if not fields else fields
        self.uri = 'http://ip-api.com/'
        self.api_key = None
        self.enable_sleep = True
        self.wait_time = datetime.now()

    def rate_limit(self, headers):
        """Method to check rate limiting. Checks ``X-Rl`` and ``X-Ttl`` headers.

        Will set timer for rate limiting if ``X-Rl`` is less than 1 for
        duration of ``X-Ttl`` + 1 second.

        Args:
            headers (dict, CaseInsensitiveDict): Request header information.

        Return:
            None
        """
        if int(headers.get('X-Rl', '0')) < 1:
            self.wait_time = datetime.now() + \
                timedelta(seconds=int(headers.get('X-Ttl', '0'))+0.25)

    def sleeper(self):
        """Method to sleep operations for rate limiting. Executes sleep.

        Return:
            None
        """
        while True:
            now = datetime.now()
            wait_time = self.wait_time - now
            if wait_time.total_seconds() < 0:
                self.wait_time = datetime.now()
                return
            wt_sec = wait_time.total_seconds()+1  # add a buffer
            logger.info(
                'Sleeping for {} seconds due to rate limiting.'.format(wt_sec))
            time.sleep(wt_sec)

    def batch(self):
        """Handle batch query operations.

        Generally not called directly, should be called by ``self.query()`` to
        allow for the logic to handle which endpoint is preferred.

        Returns:
            (list): List of resolved IP address records with specified fields.
        """
        records = []
        for ip in self.data:
            records.append({'query': ip})

        resolved_recs = []
        orig_recs = range(0, len(records), 100)
        if self.pbar:
            orig_recs = trange(0, len(records), 100,
                               desc="Resolving IPs", unit_scale=True)
        params = {
            'fields': ','.join(self.fields) if isinstance(self.fields, list) else self.fields, 'lang': self.lang,
        }
        if self.api_key:
            params['key'] = self.api_key

        for x in orig_recs:
            if self.enable_sleep:
                self.sleeper()
            rdata = requests.post(self.uri+"batch", json=records[x:x+100], params=params)

            if rdata.status_code == 200:
                self.rate_limit(rdata.headers)
                resolved_recs += rdata.json()
            elif rdata.status_code == 429:
                self.rate_limit(rdata.headers)
                self.sleeper()
                return self.batch()
            else:  # pragma: no cover
                msg = "Unknown error encountered: {}".format(rdata.status_code)
                logger.error(msg)
                resolved_recs += [{'query': result, 'status': 'failed', 'message': msg} for result in records[x:x+100]]
        return resolved_recs

    def single(self):
        """Handle single item query operations.

        Generally not called directly, should be called by ``self.query()`` to
        allow for the logic to handle which endpoint is preferred.

        Returns:
            (list): List of resolved IP address records with specified fields.
        """
        params = {
            'fields': ','.join(self.fields),
            'lang': self.lang,
        }
        if self.api_key:
            params['key'] = self.api_key

        if self.enable_sleep:
            self.sleeper()

        rdata = requests.get(
            self.uri+"json/"+self.data,
            params=params
        )
        if rdata.status_code == 200:
            self.rate_limit(rdata.headers)
            return [rdata.json()]
        if rdata.status_code == 429:
            self.rate_limit(rdata.headers)
            self.sleeper()
            return self.single()
        else:  # pragma: no cover
            msg = "Unknown error encountered: {}".format(rdata.status_code)
            logger.error(msg)
            return [{'query': self.data, 'status': 'failed', 'message': msg}]


class ProResolver(Resolver):
    """GeoIP resolver using the ip-api.com paid subscription.

    Sets endpoint to paid API, confirms support for requested language,
    and disables sleep functionality.

    Args:
        api_key (str): IP-API.com paid API key for requests.
        fields (list): Collection of fields to request in resolution.
        lang (str): Language for returned results.
    """
    def __init__(self, api_key, fields=None, lang='en'):  # pragma: no cover
        """Initialize class object and configure default values."""
        super().__init__()
        self.lang = 'en' if lang not in self.supported_langs else lang
        self.fields = [] if not fields else fields
        self.uri = 'https://pro.ip-api.com/'
        self.api_key = api_key
        self.enable_sleep = False



"""VirusTotal Resolver Tests."""
import datetime
import time
import unittest
import json
import os
from unittest.mock import patch, MagicMock

from libchickadee.resolvers.virustotal import ProResolver

__author__ = 'Chapin Bryce'
__date__ = 20200805
__license__ = 'MIT Copyright 2020 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''


class IPAPITestCase(unittest.TestCase):
    """VirusTotal Resolver Tests."""
    def setUp(self):
        """Test config"""
        self.test_data_ips = [
            '10.0.1.2', '8.8.8.8', '2001:4860:4860::8888'
        ]
        self.expected_result = [
            {'query': '10.0.1.2'},

            {'as': 'AS15169 Google LLC', 'country': 'United States',
             'org': 'Level 3', 'proxy': False, 'query': '8.8.8.8'},

            {'as': 'AS15169 Google LLC', 'country': 'United States',
             'org': 'Google LLC', 'proxy': False,
             'query': '2001:4860:4860::8888'}
        ]
        self.resolver = ProResolver(api_key='')
        local_dir = os.path.abspath(__file__).rsplit(os.sep, 1)[0]
        resource_file = open(os.path.join(local_dir, 'vt_resp_data.json'))
        self.vt_rep_data_list = json.load(resource_file)
        resource_file.close()

    def test_parse_vt_resp(self):
        """Confirm the parsing of VT responses."""
        vt_resp_data = self.vt_rep_data_list["Test0"]["test"]
        query = self.vt_rep_data_list["Test0"]["query"]
        expected = self.vt_rep_data_list["Test0"]["expected"]
        vt_values = self.resolver.parse_vt_resp(query, vt_resp_data)
        self.assertDictEqual(expected, vt_values)

    def test_parse_vt_resp_2(self):
        """Confirm the parsing of VT responses."""
        vt_resp_data = self.vt_rep_data_list["Test1"]["test"]
        query = self.vt_rep_data_list["Test1"]["query"]
        expected = self.vt_rep_data_list["Test1"]["expected"]
        vt_values = self.resolver.parse_vt_resp(query, vt_resp_data)
        self.assertDictEqual(expected, vt_values)

    @patch("libchickadee.resolvers.virustotal.requests.get")
    def test_resolve_single(self, mock_requests):
        """Test the resolution processing of a single item."""
        def mock_json():
            """Return a mocked JSON value from the request"""
            return self.vt_rep_data_list["Test0"]["test"]

        mock_requests.return_value.status_code = 200
        mock_requests.return_value.json = mock_json
        actual = self.resolver.query(self.vt_rep_data_list["Test0"]["query"])
        expected = self.vt_rep_data_list["Test0"]["expected"]
        self.assertEqual(1, len(actual))
        self.assertDictEqual(expected, actual[0])

    @patch("libchickadee.resolvers.virustotal.requests.get")
    def test_resolve_batch(self, mock_requests):
        """Test the resolution processing of a multiple items."""
        # Build 3 requests
        req1 = MagicMock()
        req1.status_code = 200
        req1.json = lambda: self.vt_rep_data_list["Test0"]["test"]

        req2 = MagicMock()
        req2.status_code = 204

        req3 = MagicMock()
        req3.status_code = 200
        req3.json = lambda: self.vt_rep_data_list["Test1"]["test"]

        mock_requests.side_effect = [req1, req2, req3]
        start = datetime.datetime.now()
        actual = self.resolver.query([self.vt_rep_data_list["Test0"]["query"],
                                      self.vt_rep_data_list["Test1"]["query"]])

        self.assertEqual(2, len(actual))
        self.assertGreaterEqual(15, (start - datetime.datetime.now()).total_seconds())
        self.assertDictEqual(self.vt_rep_data_list["Test0"]["expected"], actual[0])
        self.assertDictEqual(self.vt_rep_data_list["Test1"]["expected"], actual[1])

    @patch("libchickadee.resolvers.virustotal.requests.get")
    def test_resolve_errors(self, mock_requests):
        """Testing error handling of the resolver."""
        subtests = {
            400: "Incorrect request. Please check input data",
            403: "Authorization error. Please check API key",
            500: "Unknown error occurred, status code 500, please report"
        }
        for status_code, err_msg in subtests.items():
            mock_requests.return_value.status_code = status_code
            with self.subTest(id=status_code):
                with self.assertLogs('libchickadee.resolvers.virustotal', level='ERROR') as mock_log:
                    actual = self.resolver.query(self.vt_rep_data_list["Test0"]["query"])
                self.assertIsNone(actual)
                self.assertEqual(mock_log.records[0].message, err_msg)

    @patch("libchickadee.resolvers.virustotal.requests.get")
    def test_sleeper(self, mock_requests):
        """Validate that the sleep timer works appropriately for rate limiting"""
        initial_time = datetime.datetime.now()
        self.resolver.last_request = initial_time
        time.sleep(2)
        mock_requests.return_value.status_code = 403

        self.resolver.query(data='1.1.1.1')
        self.assertGreaterEqual(self.resolver.last_request, initial_time + datetime.timedelta(seconds=2))


if __name__ == "__main__":
    unittest.main()



"""Chickadee script tests."""
import unittest
import os
import sys
import io
from unittest.mock import patch

from libchickadee.chickadee import Chickadee, arg_handling, join_config_args, find_config_file
from libchickadee.chickadee import config_handing

__author__ = 'Chapin Bryce'
__date__ = 20200407
__license__ = 'MIT Copyright 2020 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''


class ChickadeeConfigTestCase(unittest.TestCase):
    """Test case handling configuration file functionality"""
    def setUp(self):
        """Configure defaults across tests"""
        self.default_columns = "query,message,mobile,org"
        
    def test_argparse(self):
        """Validate command line argument parsing."""
        args = [
            "1.1.1.1", "-t", "csv", "-w", "test.out", "-n", "-s",
            "--lang", "es", "-c", "test.config", "-p", "-v",
            "--log", "test.log", "-f", self.default_columns
        ]
        parsed = arg_handling(args)

        self.assertEqual(parsed.data, ["1.1.1.1"])
        self.assertEqual(parsed.fields, self.default_columns)
        self.assertEqual(parsed.output_format, "csv")
        self.assertEqual(parsed.output_file, "test.out")
        self.assertEqual(parsed.lang, "es")
        self.assertEqual(parsed.config, "test.config")
        self.assertEqual(parsed.log, "test.log")
        self.assertTrue(parsed.no_resolve)
        self.assertTrue(parsed.single)
        self.assertTrue(parsed.progress)
        self.assertTrue(parsed.verbose)

        joined = join_config_args({}, parsed)
        self.assertDictEqual(
            joined,
            {
                "data": ["1.1.1.1"],
                "fields": self.default_columns,
                "lang": "es",
                "progress": True,
                'log': 'test.log',
                'verbose': True,
                'resolver': 'ip_api',
                'virustotal': '',
                'ip_api': '',
                'no-count': False,
                'no-resolve': True,
                'include-bogon': False,
                'single': True,
                'output-format': 'csv',
                'output-file': 'test.out'
            }
        )

    def test_configparse(self):
        """Test the parsing of configuration file and command line arguments"""
        args = ["1.1.1.1", "-f", self.default_columns]
        parsed = arg_handling(args)
        opts = {
            "fields": "query,message",
            "progress": True,
            "lang": "es"
        }
        joined = join_config_args(opts, parsed)

        self.assertDictEqual(
            joined,
            {
                "data": ["1.1.1.1"],
                "fields": self.default_columns,
                "lang": "es",
                "progress": True,
                'log': os.path.abspath(os.path.join(
                    os.getcwd(), 'chickadee.log')),
                'verbose': False,
                'resolver': 'ip_api',
                'ip_api': '',
                'virustotal': '',
                'no-count': False,
                'no-resolve': False,
                'include-bogon': False,
                'single': False,
                'output-format': 'jsonl',
                'output-file': sys.stdout,
            }
        )

    def test_parse_config_file_provided(self):
        """Test the parsing of the configuration file when provided"""
        # Str, list, bool, dict, int
        test_conf = 'chickadee.ini'
        open_file = open(test_conf, 'w')
        body = """
[main]
output-format = csv
verbose = true
fields = query,country
[resolvers]
resolver = ip_api
ip_api = not-an-api-key
        """
        open_file.write(body)
        open_file.close()
        data = config_handing(test_conf)
        self.assertDictEqual(
            data,
            {
                "output-format": "csv",
                "verbose": True,
                "fields": "query,country",
                "progress": None,
                "no-resolve": None,
                'include-bogon': None,
                "log": None,
                "resolver": "ip_api",
                'virustotal': None,
                "ip_api": "not-an-api-key"
            }
        )
        os.remove(test_conf)

    def test_find_config_file(self):
        """Test the identification of a configuration file"""
        # Str, list, bool, dict, int
        test_conf = 'unittesting.chickadee.ini'
        conf_path_userdir = os.path.join(
            os.path.expanduser("~"), test_conf)
        conf_path_workdir = os.path.abspath('unittesting.chickadee.ini')
        for conf_path in [conf_path_userdir, conf_path_workdir]:
            open_conf = open(conf_path, 'w')
            open_conf.close()
            actual = find_config_file(filename_patterns=["unittesting.chickadee.ini"])
            os.remove(conf_path)
            self.assertEqual(conf_path, actual)


class ChickadeeStringTestCase(unittest.TestCase):
    """Chickadee script tests."""
    def setUp(self):
        """Set default and common values across tests"""
        self.test_data_ips = [
            '10.0.1.2', '8.8.8.8', '2001:4860:4860::8888'
        ]
        self.expected_result = [
            {'query': '10.0.1.2', 'count': 1},

            {'as': 'AS15169 Google LLC', 'country': 'United States',
             'org': 'Level 3', 'proxy': False, 'query': '8.8.8.8', 'count': 1},

            {'as': 'AS15169 Google LLC', 'country': 'United States',
             'org': 'Google LLC', 'proxy': False, 'count': 1,
             'query': '2001:4860:4860::8888'}
        ]

        self.fields = ['query', 'count', 'as', 'country', 'org', 'proxy']

    def test_no_resolve(self, resolve='No resolve'):
        """Confirm the handling when no resolution is requested"""
        results = [
            {'query': '10.0.1.2', 'count': 1, 'message': resolve},
            {'query': '8.8.8.8', 'count': 1, 'message': resolve},
            {'query': '2001:4860:4860::8888', 'count': 1,
             'message': resolve}
        ]
        for count, ip in enumerate(self.test_data_ips):
            chickadee = Chickadee()
            chickadee.fields = self.fields
            chickadee.resolve_ips = False
            chickadee.ignore_bogon = False
            data = chickadee.run(ip)
            self.assertEqual(data, [results[count]])

    @patch("libchickadee.resolvers.ipapi.Resolver.batch")
    def test_chickadee_single(self, mock_batch):
        """Query Method Test"""
        for count, ip in enumerate(self.test_data_ips):
            chickadee = Chickadee()
            chickadee.ignore_bogon = False
            chickadee.fields = self.fields
            mock_batch.return_value = [self.expected_result[count]]
            data = chickadee.run(ip)
            self.assertEqual(data, [self.expected_result[count]])

    @patch("libchickadee.resolvers.ipapi.Resolver.batch")
    def test_chickadee_csv_str(self, mock_query):
        """Batch Query Method Test"""
        chickadee = Chickadee()
        chickadee.ignore_bogon = False
        chickadee.fields = self.fields
        mock_query.return_value = self.expected_result
        data = chickadee.run(','.join(self.test_data_ips))
        self.assertCountEqual(data, self.expected_result)

    def test_chickadee_force_single(self):
        """Batch Query Method Test"""
        expected_results = self.expected_result

        class MockResolver:
            """Fake resolver"""
            def __init__(self, *args, **kwargs):
                """Set defaults"""
                self.data = None

            def single(self):
                """Mock the single method, returning a list of data."""
                return [x for x in expected_results if x['query'] == self.data]

        chickadee = Chickadee()
        chickadee.ignore_bogon = False
        chickadee.force_single = True
        chickadee.fields = self.fields
        with patch("libchickadee.chickadee.ipapi.Resolver", MockResolver):
            data = chickadee.run(','.join(self.test_data_ips))
        self.assertCountEqual(data, self.expected_result)

    def test_improper_type(self):
        """Test error handling when the wrong data type is provided via API"""
        failed = False
        try:
            # Provide improper data type
            Chickadee.str_handler(['test'])
        except TypeError:
            failed = True
        self.assertTrue(failed)

    @patch("libchickadee.resolvers.ipapi.Resolver.batch")
    def test_manual_run(self, mock_query):
        """Test the execution of chickadee via module calls"""
        chick = Chickadee(fields=self.fields)
        mock_query.return_value = [self.expected_result[1]]
        actual = chick.run(self.test_data_ips[1])
        self.assertDictEqual(self.expected_result[1], actual[0])


class ChickadeeFileTestCase(unittest.TestCase):
    """Chickadee script tests."""
    def setUp(self):
        """Test setup."""
        self.test_data_dir = os.path.join(
            os.path.dirname(__file__), 'test_data')
        google_asn = "AS15169 Google LLC"
        usa = "United States"
        google_llc = "Google LLC"
        self.txt_data_results = [
            {"as": google_asn,
             "country": usa, "org": google_llc,
             "proxy": False, "query": "2001:4860:4860::8844",
             "count": 1},

            {"as": google_asn,
             "country": usa, "org": google_llc,
             "proxy": False, "query": "2001:4860:4860::8844",
             "count": 1},

            {"query": "10.0.1.2", "count": 1},

            {"as": google_asn,
             "country": usa, "org": "Level 3",
             "proxy": False, "query": "8.8.8.8", "count": 1},

            {"as": google_asn,
             "country": usa, "org": google_llc,
             "proxy": False, "query": "2001:4860:4860::8888",
             "count": 1},

            {"as": "AS3215 Orange S.A.",
             "country": "France", "org": "", "proxy": True,
             "query": "2.2.2.2", "count": 1},

            {"as": google_asn,
             "country": usa, "org": google_llc,
             "proxy": False, "query": "2001:4860:4860::8888", "count": 1},

            {"as": "AS3356 Level 3 Parent, LLC",
             "country": usa, "org": "Informs",
             "proxy": True, "query": "4.4.4.4", "count": 1},

            {"as": "AS13335 Cloudflare, Inc.",
             "country": "Australia", "org": "", "proxy": False,
             "query": "1.1.1.1", "count": 2}
        ]

        self.fields = ['query', 'count', 'as', 'country', 'org', 'proxy']

        self.xlsx_data_results = [
            {"as": google_asn,
             "country": usa, "org": google_llc,
             "proxy": False, "query": "2001:4860:4860::8844",
             "count": 2},

            {"query": "10.0.1.2", "count": 1},

            {"as": google_asn,
             "country": usa, "org": "Level 3",
             "proxy": False, "query": "8.8.8.8", "count": 1},

            {"as": google_asn,
             "country": usa, "org": google_llc,
             "proxy": False, "query": "2001:4860:4860::8888",
             "count": 1},

            {"as": "AS3215 Orange S.A.",
             "country": "France", "org": "", "proxy": True,
             "query": "2.2.2.2", "count": 1},

            {"as": "AS3356 Level 3 Parent, LLC",
             "country": usa, "org": "Informs",
             "proxy": True, "query": "4.4.4.4", "count": 1},

            {"as": "AS13335 Cloudflare, Inc.",
             "country": "Australia", "org": "", "proxy": False,
             "query": "1.1.1.1", "count": 2}
        ]

    @patch("libchickadee.resolvers.ipapi.Resolver.batch")
    def test_ipapi_resolve_query_txt_file(self, mock_query):
        """Batch Query Method Test"""
        chickadee = Chickadee()
        chickadee.ignore_bogon = False
        chickadee.fields = self.fields
        mock_query.return_value = self.txt_data_results
        data = chickadee.run(os.path.join(self.test_data_dir, 'txt_ips.txt'))
        batch_result = []  # No reverse field
        for item in self.txt_data_results:
            batch_result.append(item)
        self.assertCountEqual(data, batch_result)

    @patch("libchickadee.resolvers.ipapi.Resolver.batch")
    def test_ipapi_resolve_query_gz_file(self, mock_query):
        """Batch Query Method Test"""
        chickadee = Chickadee()
        chickadee.ignore_bogon = False
        chickadee.fields = self.fields
        mock_query.return_value = self.txt_data_results
        data = chickadee.run(os.path.join(self.test_data_dir,
                                          'txt_ips.txt.gz'))
        batch_result = []  # No reverse field
        for item in self.txt_data_results:
            batch_result.append(item)
        self.assertCountEqual(data, batch_result)

    @patch("libchickadee.resolvers.ipapi.Resolver.batch")
    def test_ipapi_resolve_query_xlsx_file(self, mock_query):
        """Batch Query Method Test"""
        chickadee = Chickadee()
        chickadee.ignore_bogon = False
        chickadee.fields = self.fields
        mock_query.return_value = self.xlsx_data_results
        data = chickadee.run(os.path.join(self.test_data_dir, 'test_ips.xlsx'))
        batch_result = []  # No reverse field
        for item in self.xlsx_data_results:
            batch_result.append(item)
        self.assertCountEqual(data, batch_result)

    @patch("libchickadee.resolvers.ipapi.Resolver.batch")
    def test_ipapi_resolve_query_folder(self, mock_query):
        """Batch Query Method Test"""
        expected = [
            {"country": "Australia", "org": "",
             "as": "AS13335 Cloudflare, Inc.",
             "proxy": False, "query": "1.1.1.1", "count": 6},

            {"query": "10.0.1.2", "count": 3},

            {"country": "United States", "org": "Level 3",
             "as": "AS15169 Google LLC",
             "proxy": False, "query": "8.8.8.8", "count": 3},

            {"country": "United States", "org": "Google LLC",
             "as": "AS15169 Google LLC",
             "proxy": False, "query": "2001:4860:4860::8888", "count": 3},

            {"country": "United States", "org": "Informs",
             "as": "AS3356 Level 3 Parent, LLC", "proxy": True,
             "query": "4.4.4.4",
             "count": 3},

            {"country": "United States", "org": "Google LLC",
             "as": "AS15169 Google LLC",
             "proxy": False, "query": "2001:4860:4860::8844", "count": 4},

            {"country": "United States", "org": "Google LLC",
             "as": "AS15169 Google LLC",
             "proxy": False, "query": "2001:4860:4860::8888", "count": 3},

            {"country": "France", "org": "", "as": "AS3215 Orange S.A.",
             "proxy": True, "query": "2.2.2.2", "count": 3},

            {"country": "United States", "org": "Google LLC",
             "as": "AS15169 Google LLC",
             "proxy": False, "query": "2001:4860:4860::8844", "count": 4}
        ]
        mock_query.return_value = expected

        chickadee = Chickadee()
        chickadee.ignore_bogon = False
        chickadee.fields = self.fields
        data = chickadee.run(self.test_data_dir)
        self.assertCountEqual(data, expected)

    def test_file_handler_stream(self):
        """Validate the extraction of IP addresses when input is provided via stdin"""
        stream = io.TextIOWrapper(io.StringIO("test 1.1.1.1 ip"))
        ips = Chickadee.file_handler(stream, ignore_bogon=True)
        self.assertDictEqual(
            ips,
            {"1.1.1.1": 1}
        )


if __name__ == '__main__':
    unittest.main()



"""Script to check for available chickadee updates and alert the user."""

import requests
import sys

__author__ = 'Chapin Bryce'
__date__ = 20200202
__license__ = 'MIT Copyright 2020 Chapin Bryce'


def get_pypi_version():
    """Check pypi.org for version information"""
    url = 'https://pypi.org/pypi/chickadee/json'
    rdata = requests.get(url)
    return float(rdata.json().get('info', {}).get('version', '0'))


def update_available(current_version):
    """Check version against pypi.org information"""
    pypi_version = get_pypi_version()
    return pypi_version > current_version


if __name__ == "__main__":
    from libchickadee import __version__
    if update_available(__version__):
        sys.stderr.write(
            "Chickadee update is available. Please update "
            "using 'pip3 install --upgrade chickadee'.\n"
        )



"""Extract IP addresses from EVTX files."""

from libchickadee.parsers import ParserBase, run_parser_from_cli

import Evtx.Evtx


class EVTXParser(ParserBase):
    """Class to expose EVTX record contents for IP address extraction"""

    def parse_file(self, file_entry, is_stream=False):
        """Parse EVTX contents. Must be a path to an existing EVTX file.
        Cannot parse from STDIN.

        Args:
            file_entry (str): Path to EVTX file to load.
            is_stream (bool): Unused argument, required for implementation.
                Does not change functionality.
        """
        if is_stream:
            raise NotImplementedError("Providing EVTX files as an input stream of data is not yet supported.")

        # Open file
        with Evtx.Evtx.Evtx(file_entry) as event_log:
            # Iterate over events
            for record in event_log.records():
                # Send event data to self.check_ips()
                self.check_ips(record.xml())


if __name__ == "__main__":  # pragma: no cover
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('path', help="File or folder to parse")
    args = parser.parse_args()

    ev_parser = EVTXParser()
    run_parser_from_cli(args=args, parser_obj=ev_parser)




"""XLSX parsing tests."""
import unittest
import os

from libchickadee.parsers.xlsx import XLSXParser

__author__ = 'Chapin Bryce'
__date__ = 20200107
__license__ = 'MIT Copyright 2020 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''


class XLSXParserTestCase(unittest.TestCase):
    """XLSX parsing tests."""
    def setUp(self):
        """Test Config."""
        self.test_data_ips = {
            '10.0.1.2': 1,
            '8.8.8.8': 1,
            '1.1.1.1': 2,
            '2.2.2.2': 1,
            '4.4.4.4': 1,
            '2001:4860:4860::8844': 2,
            '2001:4860:4860::8888': 1
        }
        self.parser = XLSXParser(ignore_bogon=False)
        self.test_data_dir = os.path.join(
            os.path.dirname(__file__), 'test_data')

    def test_ip_extraction_xlsx(self):
        """Extraction test."""
        self.parser.parse_file(os.path.join(self.test_data_dir, 'test_ips.xlsx'))
        self.assertEqual(self.test_data_ips, self.parser.ips)


if __name__ == '__main__':
    unittest.main()



"""Plain-text parsing tests"""
import unittest

from libchickadee import __version__
from libchickadee.update import update_available, get_pypi_version

__author__ = 'Chapin Bryce'
__date__ = 20200229
__license__ = 'MIT Copyright 2020 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''


class TestUpdate(unittest.TestCase):
    """Run unit testing of the updates functionality."""
    def test_up_to_date(self):
        """Test version up to date"""
        last_public_release = get_pypi_version()
        self.assertFalse(update_available(last_public_release))

    def test_out_of_date(self):
        """Test version out of date"""
        self.assertTrue(update_available(0.0))

    def test_get_pypi_version(self):
        """Test pypi version type"""
        pypi_version = get_pypi_version()
        self.assertIsInstance(pypi_version, float)

    def test_local_version(self):
        """Test pypi version type"""
        self.assertIsInstance(__version__, float)


if __name__ == '__main__':
    unittest.main()



"""
Plain Text Parser
==================

Parse IP addresses from plain text files. Plain text files include logs,
CSVs, JSON, and other formats where ascii strings contain IPv4 or IPv6
addresses.

Also supported reading from gzipped compressed plain text data without needing
to first decompress it.

"""

import binascii
from gzip import GzipFile

from libchickadee.parsers import ParserBase, run_parser_from_cli

__author__ = 'Chapin Bryce'
__date__ = 20200114
__license__ = 'MIT Copyright 2020 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''


class PlainTextParser(ParserBase):
    """Class to extract IP addresses from plain text and gzipped plain text files."""

    @staticmethod
    def is_gz_file(filepath):
        """Validate whether the input is GZipped.

        Args:
            filepath (str): File path to test.

        Returns:
            (bool): True if a gzip file signature is identified.
        """
        with open(filepath, 'rb') as test_f:
            return binascii.hexlify(test_f.read(2)) == b'1f8b'

    def parse_file(self, file_entry, is_stream=False):
        """Parse contents of the file and extract IP addresses.

        Will read from STDIN or path to a file. Stores results in ``self.ips``.

        Args:
            file_entry (str or file_obj): Path to file for reading.
            is_stream (bool): Whether the input file is a file to open or a
                file-like object.

        Returns:
            None
        """
        if not is_stream:
            file_data = GzipFile(filename=file_entry) if self.is_gz_file(file_entry) else open(file_entry, 'rb')
        else:
            # Encode if needed
            two_bytes = file_entry.buffer.read(2)
            two_bytes = two_bytes.encode() if isinstance(two_bytes, str) else two_bytes.read(2)

            file_entry.seek(0)
            # Check for gzip stream
            file_data = GzipFile(fileobj=file_entry) if binascii.hexlify(two_bytes) == b'1f8b' else file_entry.buffer

        for raw_line in file_data:
            line = raw_line if isinstance(raw_line, str) else raw_line.decode()
            self.check_ips(line)

        if 'closed' in dir(file_data) and not file_data.closed:
            file_data.close()


if __name__ == "__main__":  # pragma: no cover
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('path', help="File or folder to parse")
    args = parser.parse_args()

    pt_parser = PlainTextParser()
    run_parser_from_cli(args=args, parser_obj=pt_parser)



"""Installer for chickadee"""
import setuptools
from libchickadee import __version__, __desc__, __author__

with open('README.md') as fh:
    long_description = fh.read()

setuptools.setup(
    name='chickadee',
    version=__version__,
    description=__desc__,
    author=__author__,
    author_email='python@chapinb.com',
    url='https://github.com/chapinb/chickadee',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': [
            'chickadee = libchickadee.chickadee:entry'
        ]
    },
    install_requires=[
        "requests ~= 2.22.0",
        "openpyxl ~= 2.6.3",
        "tqdm ~= 4.36.1",
        "netaddr ~= 0.7.19",
        "python-evtx ~= 0.6.1"
    ],
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Telecommunications Industry",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Other Audience",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Security",
        "Topic :: Utilities"
    ]
)



"""
chickadee.py
============

A command-line application to provide context for an IP address.

This utility leveraged libchickadee to extract, resolve, and report IP
addresses. Please see :ref:`installation` instructions for details
on setting up this tool on your system.

.. _chickadee-usage:

Usage
-----

.. code-block:: text

    usage: chickadee [-h] [-r {ip_api,virustotal}] [-f FIELDS]
                     [-t {json,jsonl,csv}] [-w FILENAME.JSON] [-n] [--no-count]
                     [-s] [--lang {en,de,es,pt-BR,fr,ja,zh-CN,ru}] [-b]
                     [-c CONFIG] [-p] [-v] [-V] [-l LOG]
                     [data [data ...]]

    Yet another GeoIP resolution tool.

    Will default to the free rate limited ip-api.com service for resolution.
    You can specify the paid API key for ip-api.com or for VirusTotal in
    the Chickadee configuration file. Please see template_chickadee.ini
    for more information.

    positional arguments:
      data                  Either an IP address, comma delimited list of IP addresses,
                            or path to a file or folder containing files to check for IP
                            address values. Currently supported file types: plain text
                            (ie logs, csv, json), gzipped plain text, xlsx
                            (must be xlsx extension). Can accept plain text data as
                            standard input.
                            (default: stdin)

    optional arguments:
      -h, --help            show this help message and exit
      -r {ip_api,virustotal}, --resolver {ip_api,virustotal}
                            Resolving service to use. Must specify api key in config file.
                            Please see template_chickadee.ini for instructions.
                            (default: ip_api)
      -f FIELDS, --fields FIELDS
                            Comma separated fields to query (default: None)
      -t {json,jsonl,csv}, --output-format {json,jsonl,csv}
                            Output format (default: jsonl)
      -w FILENAME.JSON, --output-file FILENAME.JSON
                            Path to file to write output
                            (default: stdout)
      -n, --no-resolve      Only extract IP addresses, don't resolve. (default: False)
      --no-count            Disable counting the occurrences of IP addresses extracted
                            from source files (default: False)
      -s, --single          Use the significantly slower single item API. Adds reverse
                            DNS. (default: False)
      --lang {en,de,es,pt-BR,fr,ja,zh-CN,ru}
                            Language (default: en)
      -b, --include-bogon   Include BOGON addresses in results. (default: False)
      -c CONFIG, --config CONFIG
                            Path to config file to load (default: None)
      -p, --progress        Enable progress bar (default: False)
      -v, --verbose         Include debug log messages (default: False)
      -V, --version         Displays version
      -l LOG, --log LOG     Path to log file (default: chickadee.log)

    Built by Chapin Bryce, v.20200805.0


.. _chickadee-examples:

chickadee Examples
----------------------

Input options
^^^^^^^^^^^^^

Parsing a single IP address:

``chickadee 1.1.1.1``

Parsing multiple IP addresses:

``chickadee 1.1.1.1,2.2.2.2``

Parsing IPs from STDIN:

``cat file.txt | chickadee``

Parsing IPs from a file:

``chickadee file.txt``

Parsing IPs from a folder, recursively:

``chickadee folder/``

Resolver options
^^^^^^^^^^^^^^^^

Resolve using VirusTotal (set API key in config file):

``chickadee -r virustotal 1.1.1.1``

Resolve using ip-api (set API key in config file):

``chickadee -r ip_api 1.1.1.1``

Output options
^^^^^^^^^^^^^^

Reporting to JSON format:

``chickadee 1.1.1.1 -t json``

Reporting to JSON lines format:

``chickadee 1.1.1.1 -t jsonl``

Reporting to CSV format:

``chickadee 1.1.1.1 -t csv``

Other Arguments
^^^^^^^^^^^^^^^

Changing the fields to resolve and report on (ip_api):

``chickadee -r ip_api -f query,count,asn,isp,org 1.1.1.1``

Changing the fields to resolve and report on (virustotal):

``chickadee -r virustotal -f query,detected_samples 1.1.1.1``

Changing the output location (STDOUT by default)

``chickadee 1.1.1.1 -w resolve.json``

Only extract IP addresses, don't resolve:

``chickadee -n 1.1.1.1``

Module Documentation
--------------------

"""

import argparse
import os
import sys
import logging
from pathlib import PurePath
from collections import Counter
import _io
import configparser

from tqdm import tqdm

# Import lib features
from libchickadee import __version__
from libchickadee.update import update_available

# Import resolvers
from libchickadee.resolvers import ipapi, virustotal, ResolverBase

# Import Parsers
from libchickadee.parsers.plain_text import PlainTextParser
from libchickadee.parsers.xlsx import XLSXParser
from libchickadee.parsers.evtx import EVTXParser


__author__ = 'Chapin Bryce'
__date__ = 20200805
__license__ = 'GPLv3 Copyright 2019 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.

Will default to the free rate limited ip-api.com service for resolution.
You can specify the paid API key for ip-api.com or for VirusTotal in
the Chickadee configuration file. Please see template_chickadee.ini
for more information.
'''

logger = logging.getLogger(__name__)


class CustomArgFormatter(argparse.RawTextHelpFormatter,
                         argparse.ArgumentDefaultsHelpFormatter):
    """Custom argparse formatter class"""


class Chickadee(object):
    """Class to handle chickadee script operations.

    Args:
        out_format (str): One of ``json``, ``jsonl``, ``csv``
        outfile (str or file_obj): Destination to write report.
        fields (list): Collection of fields to resolve and report on.

    Returns:
        None

    Examples:
        >>> chickadee = Chickadee()
        >>> resolution = chickadee.run('1.1.1.1')
        >>> print(resolution)

    """
    def __init__(self, out_format='json', outfile=sys.stdout, fields=None):
        """Initialize class values and parameters to provided or default values"""
        self.resolver = 'ip_api'
        self.input_data = None
        self.out_format = out_format
        self.outfile = outfile
        self.fields = fields
        self.force_single = False
        self.ignore_bogon = True
        self.no_count = False
        self.lang = 'en'
        self.progress_bar = False
        self.resolve_ips = True

    def run(self, input_data, api_key=None):
        """Evaluate the input data format to extract and resolve IP addresses.

        Will check the ``self.input_data`` type and select the proper handler.
        This includes handling files, directories, STDIN, and python strings
        and sending to the proper handler to extract IPs

        Once extracted, the IP addresses are passed to the ``self.resolve()``
        method if the ``self.resolve_ips`` option is enabled.

        Args:
            input_data (str or file_obj): User provided data containing IPs to
                resolve
            api_key (str): API Key for IP resolver.

        Returns:
            (list): List of dictionaries containing resolved hits.
        """
        self.input_data = input_data
        result_dict = {}
        # Extract and resolve IP addresses
        if not isinstance(self.input_data, _io.TextIOWrapper) and \
                os.path.isdir(self.input_data):
            logger.debug("Detected the data source as a directory")
            result_dict = self.dir_handler(self.input_data)  # Dir handler

        elif isinstance(self.input_data, _io.TextIOWrapper) or \
                os.path.isfile(self.input_data):
            logger.debug("Detected the data source as a file")
            # File handler
            result_dict = self.file_handler(self.input_data, self.ignore_bogon)

        elif isinstance(self.input_data, str):
            logger.debug("Detected the data source as raw value(s)")
            result_dict = self.str_handler(self.input_data)  # String handler

        logger.info("Extracted {} distinct IPs".format(
            len(list(result_dict.keys()))))

        # Resolve if requested
        if self.resolve_ips:
            return self.resolve(result_dict, api_key)

        return [{'query': k, 'count': v, 'message': 'No resolve'}
                for k, v in result_dict.items()]

    @staticmethod
    def get_api_key():
        """DEPRECIATED

        Retrieve an API key set as an environment variable. Looks for value in
        ``CHICKADEE_API_KEY``. May be depreciated in the near future.

        Returns:
            (str): API key, if found
        """
        raise NotImplementedError("Please use a configuration file to specify the API key")

    @staticmethod
    def str_handler(data):
        """Handle string input of one or more IP addresses and returns the
        distinct IPs with their associated frequency count.

        Args:
            data (list, str): raw input data from user

        Return:
            data_dict (dict): dictionary of distinct IP addresses to resolve.
        """
        if isinstance(data, str) and ',' in data:
            # List of IPs
            raw_data = data.strip().split(',')
        elif isinstance(data, str):
            # Single IP
            raw_data = [data.strip()]
        else:
            raise TypeError('Unsupported input provided.')

        # Generate a distinct list with count
        data_dict = {}
        for x in raw_data:
            if x not in data_dict:
                data_dict[x] = 0
            data_dict[x] += 1
        return data_dict

    @staticmethod
    def file_handler(file_path, ignore_bogon):
        """Handle parsing IP addresses from a file.

        Will evaluate format of input file or file stream. Currently supports
        plain text, gzipped compressed plain text, and xlsx.

        Args:
            file_path (str or file_obj): Path of file to read or stream.
            ignore_bogon (bool): Whether to include BOGON addresses in results.

        Return:
            data_dict (dict): dictionary of distinct IP addresses to resolve.
        """
        if isinstance(file_path, _io.TextIOWrapper):
            is_stream = True
            # Extract IPs with proper handler
            logger.debug("Extracting IPs from STDIN")
        else:
            is_stream = False
            # Extract IPs with proper handler
            logger.debug("Extracting IPs from {}".format(file_path))

        if not is_stream and file_path.lower().endswith('xlsx'):
            file_parser = XLSXParser(ignore_bogon)
        elif not is_stream and file_path.lower().endswith('evtx'):
            file_parser = EVTXParser(ignore_bogon)
        else:
            file_parser = PlainTextParser(ignore_bogon)
        try:
            file_parser.parse_file(file_path, is_stream)
        except Exception as e:
            logger.error("Failed to parse {}".format(file_path))
            logger.error("Error message: {}".format(e))
        return file_parser.ips

    def dir_handler(self, folder_path):
        """Handle parsing IP addresses from files recursively.

        Passes discovered files to the ``self.file_handler`` method for further
        processing.

        Args:
            folder_path (str): Directory path to recursively search for files.

        Return:
            data_dict (dict): dictionary of distinct IP addresses to resolve.
        """
        result_dict = {}
        for root, _, files in os.walk(folder_path):
            for file_name in files:
                file_entry = os.path.join(root, file_name)
                logger.debug("Parsing file {}".format(file_entry))
                file_results = self.file_handler(file_entry, self.ignore_bogon)
                logger.debug("Parsed file {}, {} results".format(
                    file_entry, len(file_results)))
                result_dict = dict(Counter(result_dict)+Counter(file_results))
        logger.debug("{} total distinct IPs discovered".format(
            len(result_dict)))
        return result_dict

    def resolve(self, data_dict, api_key=None):
        """Resolve IP addresses stored as keys within `data_dict`. The values
        for each key should represent the number of occurrences of an IP within
        a data set.

        Args:
            data_dict (dict): Structured as ``{IP: COUNT}``
            api_key (str): API Key for IP resolver.

        Returns:
            results (list): List containing resolved IP address information
        """
        distinct_ips = list(data_dict.keys())

        resolver = self.get_resolver(api_key)

        if self.progress_bar:
            resolver.pbar = self.progress_bar

        logger.debug("Resolving IPs")
        if self.force_single:
            results = []
            data = distinct_ips
            if self.progress_bar:
                data = tqdm(distinct_ips, desc="Resolving IPs", unit_scale=True)

            for element in data:
                resolver.data = element
                results += resolver.single()
        else:
            results = resolver.query(distinct_ips)

        logger.debug("Resolved IPs")

        # Add frequency information to results
        if not self.no_count:
            updated_results = []
            for result in results:
                query = str(result.get('query', ''))
                # noinspection PyTypeChecker
                result['count'] = int(data_dict.get(query, '0'))
                updated_results.append(result)

            return updated_results
        return results

    def get_resolver(self, api_key):
        """Determine the proper resolver to use, based on the available API keys.

        Args:
            api_key (str): API key value to register with the resolver

        Returns:
            Instance of an initialized resolver
        """
        resolvers = {
            "ip_api": {
                "pro_resolver": ipapi.ProResolver,
                "free_resolver": ipapi.Resolver
            },
            "virustotal": {
                "pro_resolver": virustotal.ProResolver,
                "free_resolver": None
            }
        }

        if api_key:
            logger.debug("Using authenticated resolution service")
            resolver_class = resolvers[self.resolver]['pro_resolver']
            if not resolver_class:
                raise ValueError("Unable to configure resolver. Please report to github.com/chapinb/chickadee/issues")
            resolver = resolver_class(api_key, fields=self.fields, lang=self.lang)
            logger.debug("Resolver API key found.")
        else:
            resolver_class = resolvers[self.resolver]['free_resolver']
            if not resolver_class:
                raise ValueError(
                    "Unable to configure resolver. An API key may be required for {}".format(self.resolver))
            resolver = resolver_class(fields=self.fields, lang=self.lang)

        if not self.fields:
            # Inherit the fields used by the resolver if none are used.
            self.fields = resolver.fields

        return resolver

    def write_output(self, results):
        """Write results to output format and/or files.

        Leverages the writers found in libchickadee.resolvers. Currently
        supports csv, json, and json lines formats, specified in
        ``self.out_format``.

        Args:
            results (list): List of GeoIP results

        Returns:
            None
        """

        if self.out_format == 'csv':
            logger.debug("Writing CSV report")
            ResolverBase.write_csv(self.outfile, results, self.fields)
        elif self.out_format == 'json':
            logger.debug("Writing json report")
            ResolverBase.write_json(self.outfile, results, self.fields)
        elif self.out_format == 'jsonl':
            logger.debug("Writing json lines report")
            ResolverBase.write_json(self.outfile, results, self.fields, lines=True)


def setup_logging(path, verbose=False):  # pragma: no cover
    """Function to setup logging configuration

    Args:
        path (str): File path to write log messages to
        verbose (bool): If the debug messages should be displayed on STDERR

    Returns:
        None
    """
    # Allow us to modify the `logger` variable within a function
    global logger

    # Set logger object, uses module's name
    logger = logging.getLogger(__name__)

    # Set default logger level to DEBUG. You can change this later
    logger.setLevel(logging.DEBUG)

    # Logging formatter. Best to keep consistent for most use cases
    log_format = logging.Formatter(
        '%(asctime)s %(filename)s %(levelname)s %(module)s '
        '%(funcName)s:%(lineno)d - %(message)s')

    # Setup STDERR logging, allowing you uninterrupted
    # STDOUT redirection
    stderr_handle = logging.StreamHandler(stream=sys.stderr)
    if verbose:
        stderr_handle.setLevel(logging.DEBUG)
    else:
        stderr_handle.setLevel(logging.INFO)
    stderr_handle.setFormatter(log_format)

    # Setup file logging
    file_handle = logging.FileHandler(path, 'a')
    file_handle.setLevel(logging.DEBUG)
    file_handle.setFormatter(log_format)

    # Add handles
    logger.addHandler(stderr_handle)
    logger.addHandler(file_handle)


def config_handing(config_file=None, search_conf_path=None):
    """Parse config file and return argument values.

    Args:
        config_file (str): Path to config file to read values from.
        search_conf_path (list): List of paths to look for config file

    Returns:
        dictionary containing configuration options.
    """

    # Field definitions
    # Set the default values here.
    section_defs = {
        'main': {
            'fields': '',
            'output-format': '',
            'progress': False,
            'no-resolve': False,
            'include-bogon': False,
            'log': '',
            'verbose': False
        },
        'resolvers': {
            'resolver': '',
            'ip_api': '',  # Hold respective API key
            'virustotal': ''  # Hold respective API key
        }
    }

    if not config_file:
        config_file = find_config_file(search_conf_path)

    fail_warn = 'Relying on argument defaults'
    if not config_file:
        logger.debug('Config file not found. {}'.format(fail_warn))
        return

    if not (os.path.exists(config_file) and os.path.isfile(config_file)):
        logger.debug('Error accessing config file {}. {}'.format(
            config_file, fail_warn))
        return

    conf = configparser.ConfigParser()
    conf.read(config_file)

    return parse_config_sections(conf, section_defs)


def parse_config_sections(conf, section_defs):
    """Parse the sections of the configuration file

    Args:
        conf (dict): Loaded configuration file information
        section_defs (dict): Mapping of configuration file values and defaults

    Returns:
        (dict): Final configuration to use with the script execution
    """
    config = {}
    for section, value in section_defs.items():
        if section not in conf:
            continue
        for k, v in value.items():
            conf_section = conf[section]
            conf_value = None
            if isinstance(v, str):
                conf_value = conf_section.get(k)
            elif isinstance(v, list):
                conf_value = conf_section.get(k).split(',')
            elif isinstance(v, bool):
                conf_value = conf_section.getboolean(k)
            elif isinstance(v, dict):
                raise NotImplementedError("Unable to parse dictionary objects from config file mapping")
            config[k] = conf_value
    return config


def find_config_file(search_conf_path=None, filename_patterns=None):
    """Handles the search operations for identifying configuration files on the system

    Args:
        search_conf_path (str): Path to look for a configuration file
        filename_patterns (list): Patterns to use to find a configuration file

    Returns:
        (str): The path to the first identified configuration file.
    """
    if not filename_patterns:
        # Needs to end with chickadee.ini or .chickadee.ini for detection.
        filename_patterns = ['chickadee.ini']

    if not search_conf_path:
        search_conf_path = _generate_default_config_search_path()

    for location in search_conf_path:
        if not (os.path.exists(location) and os.path.isdir(location)):
            logger.debug("Unable to access config file location {}.".format(location))
            continue
        for file_name in os.listdir(location):
            for pattern in filename_patterns:
                if file_name.endswith(pattern):
                    return os.path.join(location, file_name)


def _generate_default_config_search_path():
    """This function dynamically populates the order in which to locate a configuration file.

    Config file search path order:
      1. Current directory
      2. User home directory
      3. System wide directory

    Returns:
        (list): Ordered list of paths to look for configuration files in
    """
    search_conf_path = [os.path.abspath('.'), os.path.expanduser('~')]
    if 'win32' in sys.platform:
        search_conf_path.append(
            os.path.join(os.getenv('APPDATA'), 'chickadee'))
        search_conf_path.append('C:\\ProgramData\\chickadee')
    elif 'linux' in sys.platform or 'darwin' in sys.platform:
        search_conf_path.append(
            os.path.expanduser('~/.config/chickadee'))
        search_conf_path.append('/etc/chickadee')
    return search_conf_path


def arg_handling(args):
    """Parses command line arguments.

    Returns:
        argparse Namespace containing argument parameters.
    """
    # noinspection PyTypeChecker
    parser = argparse.ArgumentParser(
        description=__desc__,
        formatter_class=CustomArgFormatter,
        epilog="Built by {}, v.{}".format(__author__, __version__)
    )
    parser.add_argument(
        'data',
        help="Either an IP address, comma delimited list of IP addresses, "
             "or path to a file or folder containing files to check for "
             "IP address values. Currently supported file types: "
             "plain text (ie logs, csv, json), gzipped plain text, xlsx "
             "(must be xlsx extension). Can accept plain text data as stdin.",
        nargs='*',
        default=sys.stdin
    )
    parser.add_argument('-r', '--resolver',
                        help='Resolving service to use. Must specify api key in config file. '
                             'Please see template_chickadee.ini for instructions.',
                        choices=['ip_api', 'virustotal'],
                        default='ip_api')
    parser.add_argument('-f', '--fields',
                        help='Comma separated fields to query')
    parser.add_argument('-t', '--output-format',
                        help='Output format',
                        choices=['json', 'jsonl', 'csv'],
                        default='jsonl')
    parser.add_argument('-w', '--output-file',
                        help='Path to file to write output',
                        default=sys.stdout, metavar='FILENAME.JSON')
    parser.add_argument('-n', '--no-resolve', action='store_true',
                        help="Only extract IP addresses, don't resolve.")
    parser.add_argument('--no-count', action='store_true',
                        help="Disable counting the occurrences of IP addresses extracted from source files")
    parser.add_argument('-s', '--single',
                        help="Use the significantly slower single item API. "
                             "Adds reverse DNS.",
                        action='store_true')
    parser.add_argument('--lang', help="Language", default='en',
                        choices=['en', 'de', 'es', 'pt-BR', 'fr', 'ja',
                                 'zh-CN', 'ru'])
    parser.add_argument('-b', '--include-bogon', action='store_true',
                        help="Include BOGON addresses in results.")
    parser.add_argument('-c', '--config', help="Path to config file to load")
    parser.add_argument('-p', '--progress', help="Enable progress bar",
                        action="store_true")
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Include debug log messages')
    parser.add_argument('-V', '--version', action='version',
                        help='Displays version',
                        version=str(__version__))
    parser.add_argument(
        '-l',
        '--log',
        help="Path to log file",
        default=os.path.abspath(os.path.join(
            os.getcwd(), PurePath(__file__).name.rsplit('.', 1)[0] + '.log'))
    )
    return parser.parse_args(args)


def join_config_args(config, args, definitions=None):
    """Join config file and argument parameters, where the args override configs.

    Args:
        config (dict): Dictionary containing parameters from config file.
        args (obj): Argparse namespace containing command line parameters.
        definitions (dict): Dictionary of parameters to check for in args and
            config.

    Returns:
        (dict): Parameter information to use for script execution.
    """

    final_config = {}

    if not definitions:
        # This must match the defaults for argparse in order for the logic to
        # operate properly.
        definitions = {
            'fields': '',
            'output-format': 'jsonl',
            'output-file': sys.stdout,
            'progress': False,
            'no-resolve': False,
            'no-count': False,
            'include-bogon': False,
            'single': False,
            'lang': 'en',
            'log': os.path.abspath(os.path.join(
                os.getcwd(),
                PurePath(__file__).name.rsplit('.', 1)[0] + '.log')),
            'verbose': False,
            'resolver': 'ip_api',
            'ip_api': '',  # Hold the related API key
            'virustotal': '',  # Hold the related API key
            'data': ''
        }

    for k, v in definitions.items():
        args_val = getattr(args, k.replace('-', '_'), None)
        config_val = config.get(k) if config else None

        # Get from args if non-default
        if args_val != v and args_val is not None:
            final_config[k] = args_val

        # Next get from config
        elif config_val:
            final_config[k] = config_val

        # Otherwise load from args if present
        elif args_val is not None:
            final_config[k] = args_val

        # And if all else fails, load from the definitions dictionary
        else:
            final_config[k] = v

    return final_config


def entry(args=None):  # pragma: no cover
    """Entrypoint for package script.

    Args:
        args: Arguments from invocation.
    """
    if not args:
        args = sys.argv[1:]
    # Handle parameters from config file and command line.
    args = arg_handling(args)
    config = config_handing(args.config)
    params = join_config_args(config, args)

    # Check for update
    if update_available(__version__):
        sys.stderr.write(
            "Chickadee update is available. Please update "
            "using 'pip3 install --upgrade chickadee'.\n"
        )

    # Set up logging
    setup_logging(params.get('log'), params.get('verbose'))
    logger.debug("Starting Chickadee")
    for arg in vars(args):
        logger.debug("Argument {} is set to {}".format(
            arg, getattr(args, arg)
        ))

    logger.debug("Configuring Chickadee")
    fields = params.get('fields').split(',') if len(params.get('fields', [])) else None
    chickadee = Chickadee(fields=fields)
    chickadee.resolver = params.get('resolver', 'ip_api')
    chickadee.resolve_ips = not params.get('no-resolve')
    chickadee.ignore_bogon = not params.get('include-bogon')
    chickadee.no_count = params.get('no-count')
    chickadee.force_single = params.get('single')
    chickadee.lang = params.get('lang')
    chickadee.progress_bar = params.get('progress')

    logger.debug("Parsing input")
    if isinstance(params.get('data'), list):
        data = []
        for x in params.get('data'):
            res = chickadee.run(x, params.get(chickadee.resolver))
            data += res
    else:
        data = chickadee.run(params.get('data'), params.get(chickadee.resolver))

    logger.debug("Writing output")
    chickadee.outfile = params.get('output-file')
    chickadee.out_format = params.get('output-format')
    chickadee.write_output(data)

    logger.debug("Chickadee complete")


if __name__ == "__main__":
    entry()



"""
XLSX File Parser
==================

Parse IP addresses from XLSX files. This will extract IP addresses stored as
values (not formulas) across all tabs within a spreadsheet.

"""

from openpyxl import load_workbook

from libchickadee.parsers import ParserBase, run_parser_from_cli

__author__ = 'Chapin Bryce'
__date__ = 20200107
__license__ = 'MIT Copyright 2020 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''


class XLSXParser(ParserBase):
    """Class to extract IP addresses from XLSX workbooks."""

    def parse_file(self, file_entry, is_stream=False):
        """Parse xlsx contents. Must be a path to an existing XLSX workbook.
        Cannot parse from STDIN.

        Args:
            file_entry (str): Path to workbook to load.
            is_stream (bool): Unused argument, required for implementation.
                Does not change functionality.
        """
        if is_stream:
            raise NotImplementedError("Providing XLSX files as an input stream of data is not yet supported.")

        wb = load_workbook(file_entry)

        for sheet in wb.sheetnames:
            ws = wb[sheet]
            for row in ws.iter_rows():
                for cell in row:
                    if isinstance(cell.value, (str, bytes)):
                        self.check_ips(cell.value)


if __name__ == '__main__':  # pragma: no cover
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('path', help="File or folder to parse")
    args = parser.parse_args()

    xl_parser = XLSXParser()
    run_parser_from_cli(args=args, parser_obj=xl_parser)
